from numbers import Number
from stats_arrays import (
    UndefinedUncertainty,
    UniformUncertainty,
    TriangularUncertainty,
    NormalUncertainty,
    NoUncertainty,
    LognormalUncertainty,
)


def get_uncertainty_type(obj):
    guesses = ("uncertainty type", "uncertainty_type", "uncertainty_type_id")
    for guess in guesses:
        if guess in obj:
            return obj[guess]
    return 0


def rescale_object(obj, factor):
    """Rescale objects, including formulas and uncertainty values, by a constant factor.

    """
    if not isinstance(factor, Number):
        raise ValueError("Must provide a number for `factor`")
    ut = get_uncertainty_type(obj)
    if obj.get("formula"):
        obj["formula"] = "({}) * {}".format(obj["formula"], factor)
    if ut in (UndefinedUncertainty.id, NoUncertainty.id):
        obj["amount"] = factor * obj["amount"]
        if "loc" in obj:
            obj["loc"] = obj["amount"]
    elif ut == NormalUncertainty.id:
        obj["amount"] = obj["loc"] = factor * obj["amount"]
        obj["scale"] *= factor
    elif ut == LognormalUncertainty.id:
        # ``scale`` in lognormal is scale-independent
        obj["amount"] = obj["loc"] = factor * obj["amount"]
    elif ut == TriangularUncertainty.id:
        if factor < 0:
            obj["minimum"], obj["maximum"] = (
                factor * obj["maximum"],
                factor * obj["minimum"],
            )
        else:
            obj["minimum"] *= factor
            obj["maximum"] *= factor
        obj["amount"] = obj["loc"] = factor * obj["amount"]
    elif ut == UniformUncertainty.id:
        if factor < 0:
            obj["minimum"], obj["maximum"] = (
                factor * obj["maximum"],
                factor * obj["minimum"],
            )
        else:
            obj["minimum"] *= factor
            obj["maximum"] *= factor
        if "amount" in obj:
            obj["amount"] *= factor
    else:
        raise ValueError("This object type can't be automatically rescaled")
    return obj



from .version import version as __version__

__all__ = (
    "get_migration",
    "migrate_data",
    "load_and_clean_exiobase_3_ecoinvent_36_migration",
)

import pandas as pd
from bw2data import Database
from .strategies import migrate_data, get_migration


def load_and_clean_exiobase_3_ecoinvent_36_migration(
    ecoinvent_biosphere_name=None, explode=False
):

    # load migration data
    migration = get_migration("exiobase-3-ecoinvent-3.6")

    # convert migration data into dataframe
    df = pd.DataFrame(
        [
            {"exiobase name": k[0], "exiobase compartment": k[1], "value": v}
            if isinstance(v, list)
            else {"exiobase name": k[0], "exiobase compartment": k[1], "value": [v]}
            for k, v in migration["data"]
        ]
    )

    # convert lists into individual rows
    df = df.explode("value").reset_index(drop=True)

    # convert dicts into individual columns
    df = df.join(df["value"].apply(pd.Series)).drop(columns=["value"])

    # rename columns
    df = df.rename(
        columns={
            "unit": "ecoinvent unit",
            "name": "ecoinvent name",
            "categories": "ecoinvent categories",
            "__multiplier__": "multiplier",
            "__disaggregation__": "disaggregation",
        }
    )

    # fill in missing data
    df = df.fillna(
        {
            "exiobase compartment": "undef",
            "ecoinvent name": df["exiobase name"],
            "disaggregation": 1,
            "ecoinvent categories": df["exiobase compartment"],
        }
    )

    # convert ecoinvent categories into tuples
    df.loc[:, "ecoinvent categories"] = df["ecoinvent categories"].apply(
        lambda x: tuple(x) if isinstance(x, list) else (x,)
    )

    # combine multiplier and disaggregation into one 'factor' column
    df["factor"] = df["multiplier"] * df["disaggregation"]

    # merge with ecoinvent biosphere to extract global flow indices
    if ecoinvent_biosphere_name is not None:
        df["biosphere index"] = df.merge(
            pd.DataFrame(
                Database(ecoinvent_biosphere_name)
            ),  # get ecoinvent biosphere as dataframe
            how="left",  # keep left index
            left_on=[
                "ecoinvent name",
                "ecoinvent categories",
                "ecoinvent unit",
            ],  # columns to merge on
            right_on=["name", "categories", "unit"],  # columns to merge on
        )["id"]

    # undo exploding for unique index
    if explode is False:
        df = df.groupby(["exiobase name", "exiobase compartment"], dropna=False).agg(
            list
        )

    return df


# def create_core_migrations():
#     """Add pre-defined core migrations data files"""
#     Migration("biosphere-2-3-categories").write(
#         get_biosphere_2_3_category_migration_data(),
#         "Change biosphere category and subcategory labels to ecoinvent version 3"
#     )
#     Migration("biosphere-2-3-names").write(
#         get_biosphere_2_3_name_migration_data(),
#         "Change biosphere flow names to ecoinvent version 3"
#     )
#     Migration("simapro-ecoinvent-3.1").write(
#         get_simapro_ecoinvent_3_migration_data("3.1"),
#         "Change SimaPro names from ecoinvent 3.1 to ecoinvent names"
#     )
#     Migration("simapro-ecoinvent-3.2").write(
#         get_simapro_ecoinvent_3_migration_data("3.2"),
#         "Change SimaPro names from ecoinvent 3.2 to ecoinvent names"
#     )
#     Migration("simapro-ecoinvent-3.3").write(
#         get_simapro_ecoinvent_3_migration_data("3.3"),
#         "Change SimaPro names from ecoinvent 3.3 to ecoinvent names"
#     )
#     Migration("simapro-water").write(
#         get_simapro_water_migration_data(),
#         "Change SimaPro water flows to more standard names"
#     )
#     Migration("us-lci").write(
#         get_us_lci_migration_data(),
#         "Fix names in US LCI database"
#     )
#     Migration("default-units").write(
#         get_default_units_migration_data(),
#         "Convert to default units"
#     )
#     Migration("unusual-units").write(
#         get_unusual_units_migration_data(),
#         "Convert non-Ecoinvent units"
#     )
#     Migration("exiobase-biosphere").write(
#         get_exiobase_biosphere_migration_data(),
#         "Change biosphere flow names to ecoinvent version 3"
#     )
#     Migration("fix-ecoinvent-flows-pre-35").write(
#         get_ecoinvent_pre35_migration_data(),
#         "Update new biosphere UUIDs in Consequential 3.4"
#     )



from .utils import rescale_object
from collections.abc import Mapping
from copy import deepcopy
from pathlib import Path
import json

DATA_DIR = Path(__file__, "..").resolve() / "data"


def get_migration(location):
    if isinstance(location, Mapping):
        return location
    elif isinstance(location, Path) and Path(location).is_file():
        return json.load(open(location))
    elif isinstance(location, str) and Path(location).is_file():
        return json.load(open(location))
    elif isinstance(location, str) and any((location + ".json") in o.name for o in DATA_DIR.iterdir()):
        return json.load(open(DATA_DIR / (location + ".json")))
    else:
        raise ValueError


def modify_object(obj, dct):
    scale = dct.get("__disaggregation__", 1) * dct.get("__multiplier__", 1)
    if scale != 1:
        if 'amount' not in obj:
            raise ValueError("Rescale needed but not `amount` field present")
        obj = rescale_object(obj, scale)
    for k, v in dct.items():
        if k not in ("__disaggregation__", "__multiplier__"):
            obj[k] = v
    return obj


def migrate_data(data, migration):
    migration_data = get_migration(migration)
    lookup = {tuple(x): y for x, y in migration_data["data"]}

    for row in data:
        try:
            new = lookup[tuple([row.get(field) for field in migration_data["fields"]])]
            if isinstance(new, list):
                for dct in new:
                    yield modify_object(deepcopy(row), dct)
            else:
                yield modify_object(row, new)
        except KeyError:
            yield row



from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))

requirements = ["stats_arrays", "bw2data", "pandas"]
test_requirements = ["pytest"]

v_temp = {}
with open("bw_migrations/version.py") as fp:
    exec(fp.read(), v_temp)
version = ".".join((str(x) for x in v_temp["version"]))


setup(
    name="bw_migrations",
    version=version,
    description="Migration data and utilities for Brightway IO and LCA in general",
    long_description=open(path.join(here, "README.md")).read(),
    long_description_content_type="text/markdown",
    url="https://github.com/brightway-lca/bw_migrations",
    author="Chris Mutel",
    author_email="cmutel@gmail.com",
    license="NewBSD 3-clause; LICENSE",
    package_data={'bw_migrations': ["data/*.json"]},
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 3.7",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Mathematics",
    ],
    packages=find_packages(exclude=["tests"]),
    install_requires=requirements,
    tests_require=requirements + test_requirements,
)
